from sys import version_info

from panther.db.connection import db
from panther.db.utils import merge_dicts, prepare_id_for_query

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BasePantherDBQuery')


class BasePantherDBQuery:
    @classmethod
    def _merge(cls, *args) -> dict:
        prepare_id_for_query(*args)
        return merge_dicts(*args)

    # # # # # Find # # # # #
    @classmethod
    def find_one(cls, _data: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).find_one(**cls._merge(_data, kwargs)):
            return cls.create_model_instance(document=document)
        return None

    @classmethod
    def find(cls, _data: dict | None = None, /, **kwargs) -> list[Self]:
        documents = db.session.collection(cls.__name__).find(**cls._merge(_data, kwargs))
        return [cls.create_model_instance(document=document) for document in documents]

    @classmethod
    def first(cls, _data: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).first(**cls._merge(_data, kwargs)):
            return cls.create_model_instance(document=document)
        return None

    @classmethod
    def last(cls, _data: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).last(**cls._merge(_data, kwargs)):
            return cls.create_model_instance(document=document)
        return None

    # # # # # Count # # # # #
    @classmethod
    def count(cls, _data: dict | None = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).count(**cls._merge(_data, kwargs))

    # # # # # Insert # # # # #
    @classmethod
    def insert_one(cls, _data: dict | None = None, **kwargs) -> Self:
        document = db.session.collection(cls.__name__).insert_one(**cls._merge(_data, kwargs))
        return cls.create_model_instance(document=document)

    # # # # # Delete # # # # #
    def delete(self) -> None:
        db.session.collection(self.__class__.__name__).delete_one(_id=self.id)

    @classmethod
    def delete_one(cls, _data: dict | None = None, /, **kwargs) -> bool:
        return db.session.collection(cls.__name__).delete_one(**cls._merge(_data, kwargs))

    @classmethod
    def delete_many(cls, _data: dict | None = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).delete_many(**cls._merge(_data, kwargs))

    # # # # # Update # # # # #
    def update(self, **kwargs) -> None:
        for field, value in kwargs.items():
            setattr(self, field, value)
        db.session.collection(self.__class__.__name__).update_one({'_id': self.id}, **kwargs)

    @classmethod
    def update_one(cls, _filter: dict, _data: dict | None = None, /, **kwargs) -> bool:
        prepare_id_for_query(_filter)
        return db.session.collection(cls.__name__).update_one(_filter, **cls._merge(_data, kwargs))

    @classmethod
    def update_many(cls, _filter: dict, _data: dict | None = None, /, **kwargs) -> int:
        prepare_id_for_query(_filter)
        return db.session.collection(cls.__name__).update_many(_filter, **cls._merge(_data, kwargs))



import contextlib
import logging
import sys
import types
from collections.abc import Callable
from logging.config import dictConfig
from pathlib import Path
from threading import Thread

import panther.logging
from panther import status
from panther._load_configs import *
from panther._utils import clean_traceback_message, http_response, run_sync_async_function
from panther.background_tasks import background_tasks
from panther.cli.utils import print_info
from panther.configs import config
from panther.exceptions import APIException, PantherException
from panther.monitoring import Monitoring
from panther.request import Request
from panther.response import Response
from panther.routings import collect_path_variables, find_endpoint


dictConfig(panther.logging.LOGGING)
logger = logging.getLogger('panther')


class Panther:
    def __init__(self, name: str, configs=None, urls: dict | None = None, startup: Callable = None, shutdown: Callable = None):
        self._configs = configs
        self._urls = urls
        self._startup = startup
        self._shutdown = shutdown

        config['base_dir'] = Path(name).resolve().parent

        try:
            self.load_configs()
        except Exception as e:  # noqa: BLE001
            if isinstance(e, PantherException):
                logger.error(e.args[0])
            else:
                logger.error(clean_traceback_message(e))
            sys.exit()

        # Monitoring
        self.monitoring = Monitoring(is_active=config['monitoring'])

        # Print Info
        print_info(config)

        # Startup
        self.handle_startup()

        # Start Websocket Listener (Redis Required)
        if config['has_ws']:
            Thread(
                target=self.websocket_connections,
                daemon=True,
                args=(self.ws_redis_connection,),
            ).start()

    def load_configs(self) -> None:

        # Check & Read The Configs File
        self.configs = load_configs_file(self._configs)

        # Put Variables In "config" (Careful about the ordering)
        config['secret_key'] = load_secret_key(self.configs)
        config['monitoring'] = load_monitoring(self.configs)
        config['log_queries'] = load_log_queries(self.configs)
        config['background_tasks'] = load_background_tasks(self.configs)
        config['throttling'] = load_throttling(self.configs)
        config['default_cache_exp'] = load_default_cache_exp(self.configs)
        config['middlewares'] = load_middlewares(self.configs)
        config['reversed_middlewares'] = config['middlewares'][::-1]
        config['user_model'] = load_user_model(self.configs)
        config['authentication'] = load_authentication_class(self.configs)
        config['jwt_config'] = load_jwt_config(self.configs)
        config['startup'] = load_startup(self.configs)
        config['shutdown'] = load_shutdown(self.configs)
        config['models'] = collect_all_models()

        # Initialize Background Tasks
        if config['background_tasks']:
            background_tasks.initialize()

        # Load URLs should be one of the last calls in load_configs,
        #   because it will read all files and loads them.
        config['flat_urls'], config['urls'] = load_urls(self.configs, urls=self._urls)
        config['urls']['_panel'] = load_panel_urls()

        self._create_ws_connections_instance()

    def _create_ws_connections_instance(self):
        from panther.base_websocket import Websocket
        from panther.websocket import WebsocketConnections

        # Check do we have ws endpoint
        for endpoint in config['flat_urls'].values():
            if not isinstance(endpoint, types.FunctionType) and issubclass(endpoint, Websocket):
                config['has_ws'] = True
                break
        else:
            config['has_ws'] = False

        # Create websocket connections instance
        if config['has_ws']:
            config['websocket_connections'] = self.websocket_connections = WebsocketConnections()
            # Websocket Redis Connection
            for middleware in config['middlewares']:
                if middleware.__class__.__name__ == 'RedisMiddleware':
                    self.ws_redis_connection = middleware.redis_connection_for_ws()
                    break
            else:
                self.ws_redis_connection = None

    async def __call__(self, scope: dict, receive: Callable, send: Callable) -> None:
        """
        1.
            await func(scope, receive, send)
        2.
            async with asyncio.TaskGroup() as tg:
                tg.create_task(func(scope, receive, send))
        3.
            async with anyio.create_task_group() as task_group:
                task_group.start_soon(func, scope, receive, send)
                await anyio.to_thread.run_sync(func, scope, receive, send)
        4.
            with ProcessPoolExecutor() as e:
                e.submit(func, scope, receive, send)
        """
        if scope['type'] == 'lifespan':
            return
        func = self.handle_http if scope['type'] == 'http' else self.handle_ws
        await func(scope=scope, receive=receive, send=send)

    async def handle_ws(self, scope: dict, receive: Callable, send: Callable) -> None:
        from panther.websocket import GenericWebsocket, Websocket
        monitoring = Monitoring(is_active=config['monitoring'], is_ws=True)

        temp_connection = Websocket(scope=scope, receive=receive, send=send)
        await monitoring.before(request=temp_connection)

        endpoint, found_path = find_endpoint(path=temp_connection.path)
        if endpoint is None:
            await monitoring.after('Rejected')
            return await temp_connection.close(status.WS_1000_NORMAL_CLOSURE)
        path_variables: dict = collect_path_variables(request_path=temp_connection.path, found_path=found_path)

        if not issubclass(endpoint, GenericWebsocket):
            logger.critical(f'You may have forgotten to inherit from GenericWebsocket on the {endpoint.__name__}()')
            await monitoring.after('Rejected')
            return await temp_connection.close(status.WS_1014_BAD_GATEWAY)

        del temp_connection
        connection = endpoint(scope=scope, receive=receive, send=send)
        connection.set_path_variables(path_variables=path_variables)

        # Call 'Before' Middlewares
        for middleware in config['middlewares']:
            try:
                connection = await middleware.before(request=connection)
            except APIException:
                await connection.close()
                break
        else:
            await self.websocket_connections.new_connection(connection=connection)
            await monitoring.after('Accepted')
            await connection.listen()

        # Call 'After' Middleware
        for middleware in config['reversed_middlewares']:
            with contextlib.suppress(APIException):
                await middleware.after(response=connection)

        await monitoring.after('Closed')
        return None

    async def handle_http(self, scope: dict, receive: Callable, send: Callable) -> None:
        request = Request(scope=scope, receive=receive, send=send)

        # Monitoring
        await self.monitoring.before(request=request)

        # Read Request Payload
        await request.read_body()

        # Find Endpoint
        endpoint, found_path = find_endpoint(path=request.path)
        path_variables: dict = collect_path_variables(request_path=request.path, found_path=found_path)

        if endpoint is None:
            return await http_response(
                send,
                status_code=status.HTTP_404_NOT_FOUND,
                monitoring=self.monitoring,
                exception=True,
            )

        try:  # They Both(middleware.before() & _endpoint()) Have The Same Exception (APIException)
            # Call 'Before' Middlewares
            for middleware in config['middlewares']:
                request = await middleware.before(request=request)

            # Function
            if isinstance(endpoint, types.FunctionType):
                # Function Doesn't Have @API Decorator
                if not hasattr(endpoint, '__wrapped__'):
                    logger.critical(f'You may have forgotten to use @API() on the {endpoint.__name__}()')
                    return await http_response(
                        send,
                        status_code=status.HTTP_501_NOT_IMPLEMENTED,
                        monitoring=self.monitoring,
                        exception=True,
                    )

                # Declare Endpoint
                _endpoint = endpoint

            # Class
            else:
                from panther.app import GenericAPI

                if not issubclass(endpoint, GenericAPI):
                    logger.critical(f'You may have forgotten to inherit from GenericAPI on the {endpoint.__name__}()')
                    return await http_response(
                        send,
                        status_code=status.HTTP_501_NOT_IMPLEMENTED,
                        monitoring=self.monitoring,
                        exception=True,
                    )
                # Declare Endpoint
                _endpoint = endpoint.call_method

            # Call Endpoint
            response = await _endpoint(request=request, **path_variables)

        except APIException as e:
            response = self.handle_exceptions(e)
        except Exception as e:  # noqa: BLE001
            # Every unhandled exception in Panther or code will catch here
            exception = clean_traceback_message(exception=e)
            logger.critical(exception)

            return await http_response(
                send,
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                monitoring=self.monitoring,
                exception=True,
            )

        # Call 'After' Middleware
        for middleware in config['reversed_middlewares']:
            try:
                response = await middleware.after(response=response)
            except APIException as e:  # noqa: PERF203
                response = self.handle_exceptions(e)

        await http_response(
            send,
            status_code=response.status_code,
            monitoring=self.monitoring,
            headers=response.headers,
            body=response.body,
        )

    def handle_startup(self):
        if startup := config['startup'] or self._startup:
            run_sync_async_function(startup)

    def handle_shutdown(self):
        if shutdown := config['shutdown'] or self._shutdown:
            run_sync_async_function(shutdown)

    def __del__(self):
        self.handle_shutdown()

    @classmethod
    def handle_exceptions(cls, e: APIException, /) -> Response:
        return Response(
            data=e.detail if isinstance(e.detail, dict) else {'detail': e.detail},
            status_code=e.status_code,
        )



import sys
from unittest import TestCase

from panther import Panther
from panther.test import APIClient


class TestSimpleRequests(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/app')
        from tests.app.urls import simple_requests_urls

        app = Panther(__name__, configs=__name__, urls=simple_requests_urls)
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        sys.path.pop()

    def test_header(self):
        res = self.client.get('request-header/')
        assert res.status_code == 200
        headers = [
            'accept_encoding',
            'content_length',
            'authorization',
            'content_type',
            'user_agent',
            'connection',
            'accept',
            'host',
            'sec_websocket_version',
            'sec_websocket_key',
            'upgrade',
        ]
        assert [*res.data.keys()] == headers
        assert not all(res.data.values())

    def test_path(self):
        res = self.client.get('request-path/')
        assert res.status_code == 200
        assert res.data == '/request-path/'

    def test_client(self):
        res = self.client.get('request-client/')
        assert res.status_code == 200
        assert res.data == ['127.0.0.1', 8585]

    def test_query_params(self):
        res = self.client.get(
            'request-query_params/',
            query_params={'my': 'name', 'is': 'ali', 'how': 'are'},
        )
        assert res.status_code == 200
        assert res.data == {'my': 'name', 'is': 'ali', 'how': 'are'}



import logging
import re
from collections import Counter
from collections.abc import Callable, Mapping, MutableMapping
from copy import deepcopy
from functools import partial, reduce

from panther.configs import config


logger = logging.getLogger('panther')


def flatten_urls(urls: dict) -> dict:
    return dict(_flattening_urls(urls))


def _flattening_urls(data: dict | Callable, url: str = ''):
    # Add `/` add the end of url
    if not url.endswith('/'):
        url = f'{url}/'

    if isinstance(data, dict):
        for k, v in data.items():
            yield from _flattening_urls(v, f'{url}{k}')
    else:
        # Remove `/` prefix of url
        url = url.removeprefix('/')

        # Collect it, if it doesn't have problem
        if _is_url_endpoint_valid(url=url, endpoint=data):
            yield url, data


def _is_url_endpoint_valid(url: str, endpoint: Callable) -> bool:
    if endpoint is ...:
        logger.error(f"URL Can't Point To Ellipsis. ('{url}' -> ...)")
    elif endpoint is None:
        logger.error(f"URL Can't Point To None. ('{url}' -> None)")
    elif url and not re.match(r'^[a-zA-Z<>0-9_/-]+$', url):
        logger.error(f"URL Is Not Valid. --> '{url}'")
    else:
        return True
    return False


def finalize_urls(urls: dict) -> dict:
    """Convert flat dict to nested"""
    urls_list = []
    for url, endpoint in urls.items():
        path = {}
        if url == '':
            # This condition only happen when
            #   user defines the root url == '' instead of '/'
            url = '/'

        for single_path in list(filter(lambda x: x != '', url.split('/')[:-1][::-1])) or ['']:
            if single_path != '' and not path:
                path = {single_path: {'': endpoint}}
            else:
                path = {single_path: path or endpoint}
        urls_list.append(path)
    return _merge(*urls_list) if urls_list else {}


def _merge(destination: MutableMapping, *sources) -> MutableMapping:
    return _simplify_urls(reduce(partial(_deepmerge), sources, destination))


def _simplify_urls(urls):
    simplified_urls = {}

    for key, value in urls.items():
        if isinstance(value, dict):
            simplified_value = _simplify_urls(value)
            if isinstance(simplified_value, dict) and len(simplified_value) == 1 and '' in simplified_value:
                simplified_urls[key] = simplified_value['']
            else:
                simplified_urls[key] = simplified_value
        else:
            simplified_urls[key] = value

    return simplified_urls


def _deepmerge(dst, src):
    """Credit to Travis Clarke --> https://github.com/clarketm/mergedeep"""
    for key in src:
        if key in dst:
            if _is_recursive_merge(dst[key], src[key]):
                _deepmerge(dst[key], src[key])
            else:
                dst[key] = deepcopy(src[key])
        else:
            dst[key] = deepcopy(src[key])
    return dst


def _is_recursive_merge(a, b):
    both_mapping = isinstance(a, Mapping) and isinstance(b, Mapping)
    both_counter = isinstance(a, Counter) and isinstance(b, Counter)
    return both_mapping and not both_counter


def find_endpoint(path: str) -> tuple[Callable | None, str]:
    urls = config['urls']

    if (location := path.find('?')) != -1:
        path = path[:location]
    path = path.removesuffix('/').removeprefix('/')  # 'user/list'
    paths = path.split('/')  # ['user', 'list']
    paths_len = len(paths)

    found_path = ''
    for i, split_path in enumerate(paths):
        last_path = bool((i + 1) == paths_len)
        found = urls.get(split_path)

        # `found` is callable
        if last_path and callable(found):
            found_path += f'{split_path}/'
            return found, found_path

        # `found` is dict
        if isinstance(found, dict):
            found_path += f'{split_path}/'
            if last_path and callable(endpoint := found.get('')):
                return endpoint, found_path

            urls = found
            continue

        # `found` is None
        _continue = False
        for key, value in urls.items():
            if key.startswith('<'):
                if last_path:
                    if callable(value):
                        found_path += f'{key}/'
                        return value, found_path

                    if isinstance(value, dict) and '' in value:
                        found_path += f'{key}/'
                        return value[''], found_path

                    return None, ''

                urls = value
                found_path += f'{key}/'
                _continue = True
                break

        if _continue:
            continue

    return None, ''


def collect_path_variables(request_path: str, found_path: str) -> dict:
    found_path = found_path.removesuffix('/').removeprefix('/')
    request_path = request_path.removesuffix('/').removeprefix('/')
    path_variables = {}
    for f_path, r_path in zip(found_path.split('/'), request_path.split('/')):
        if f_path.startswith('<'):
            path_variables[f_path[1:-1]] = r_path
    return path_variables



from datetime import timedelta
from pathlib import Path
from typing import TypedDict, Callable

from pydantic._internal._model_construction import ModelMetaclass

from panther.throttling import Throttling


class JWTConfig:
    def __init__(
            self,
            key: str,
            algorithm: str = 'HS256',
            life_time: timedelta | int = timedelta(days=1),
            refresh_life_time: timedelta | int | None = None,
    ):
        self.key = key
        self.algorithm = algorithm
        self.life_time = life_time.total_seconds() if isinstance(life_time, timedelta) else life_time

        if refresh_life_time:
            if isinstance(refresh_life_time, timedelta):
                self.refresh_life_time = refresh_life_time.total_seconds()
            else:
                self.refresh_life_time = refresh_life_time
        else:
            self.refresh_life_time = self.life_time * 2


class Config(TypedDict):
    base_dir: Path
    monitoring: bool
    log_queries: bool
    default_cache_exp: timedelta | None
    throttling: Throttling | None
    secret_key: bytes | None
    middlewares: list
    reversed_middlewares: list
    user_model: ModelMetaclass | None
    authentication: ModelMetaclass | None
    jwt_config: JWTConfig | None
    models: list[dict]
    flat_urls: dict
    urls: dict
    db_engine: str
    websocket_connections: any  # type: WebsocketConnections
    background_tasks: bool
    has_ws: bool
    startup: Callable
    shutdown: Callable


config: Config = {
    'base_dir': Path(),
    'monitoring': False,
    'log_queries': False,
    'default_cache_exp': None,
    'throttling': None,
    'secret_key': None,
    'middlewares': [],
    'reversed_middlewares': [],
    'user_model': None,
    'authentication': None,
    'jwt_config': None,
    'models': [],
    'flat_urls': {},
    'urls': {},
    'db_engine': '',
    'websocket_connections': None,
    'background_tasks': False,
    'has_ws': False,
    'startup': None,
    'shutdown': None,
}



import logging
from rich import print as rprint


logger = logging.getLogger('panther')


logo = r"""│    ____                 __    __                         │
│   /\  _`\              /\ \__/\ \                        │
│   \ \ \L\ \ __      ___\ \ ,_\ \ \___      __   _ __     │
│    \ \ ,__/'__`\  /' _ `\ \ \/\ \  _ `\  /'__`\/\`'__\   │
│     \ \ \/\ \L\.\_/\ \/\ \ \ \_\ \ \ \ \/\  __/\ \ \/    │
│      \ \_\ \__/.\_\ \_\ \_\ \__\\ \_\ \_\ \____\\ \_\    │
│       \/_/\/__/\/_/\/_/\/_/\/__/ \/_/\/_/\/____/ \/_/    │
"""

help_message = f"""╭{58 * '─'}╮
{logo}│{58 * ' '}│
│                                                          │
│   usage:                                                 │
│       - panther create <project_name> <directory>        │
│           Create project in <directory> default is `.`   │
│                                                          │
│       - panther run [--reload | --help]                  │
│           Run your project with uvicorn                  │
│                                                          │
│       - panther shell [ bpython | ipython ]              │
│           Run interactive python shell                   │
│                                                          │
│       - panther monitor                                  │
│           Show the monitor :)                            │
│                                                          │
│       - panther version                                  │
│           Print the current version of Panther           │
│                                                          │
│       - panther [--help | -h ]                           │
│           Show this message and exit                     │
╰{58 * '─'}╯
"""


def cli_error(message: str | Exception) -> None:
    logger.error(message)
    logger.error('Use "panther -h" for more help')


def cli_warning(message: str | Exception, hint: str = None) -> None:
    logger.warning(message)
    if hint:
        logger.info(hint)


def cli_info(message: str) -> None:
    logger.info(message)
    logger.info('Use "panther -h" for more help\n')


def clean_args(args: list[str]) -> dict:
    """
    Input: ['--reload', '--host', '127.0.0.1', ...]
    Output: {'--reload: None, 'host': '127.0.0.1', ...}
    """
    _args = {}
    for i, arg in enumerate(args):
        if arg.startswith('--'):
            if (i + 1) < len(args):
                _args[arg[2:]] = args[i + 1]
            else:
                _args[arg[2:]] = True
    return _args


def print_help_message():
    rprint(help_message)


def print_uvicorn_help_message():
    rprint('Run `uvicorn --help` for more help')


def print_info(config: dict):
    mo = config['monitoring']
    lq = config['log_queries']
    bt = config['background_tasks']
    ws = config['has_ws']
    bd = '{0:<39}'.format(str(config['base_dir']))
    if len(bd) > 39:
        bd = f'{bd[:36]}...'

    if config['monitoring']:
        monitor = '│ * Run "panther monitor" in another session for Monitoring│'
    else:
        monitor = f'│{58 * " "}│'

    info_message = f"""
╭{58 * '─'}╮
{logo}│{58 * ' '}│
│   Monitoring: {mo}                                  \t   │
│   Log Queries: {lq}                                 \t   │
│   Background Tasks: {bt}                            \t   │
│   Websocket: {ws}                                   \t   │
│   Base directory: {bd}│
{monitor}
╰{58 * '─'}╯
"""
    rprint(info_message)



from types import NoneType

import orjson as json
from pydantic import BaseModel as PydanticBaseModel
from pydantic._internal._model_construction import ModelMetaclass

ResponseDataTypes = list | tuple | set | dict | int | str | bool | bytes | NoneType | ModelMetaclass
IterableDataTypes = list | tuple | set


class Response:
    content_type = 'application/json'

    def __init__(
        self,
        data: ResponseDataTypes = None,
        headers: dict | None = None,
        status_code: int = 200,
    ):
        """
        :param data: should be int | dict | list | tuple | set | str | bool | bytes | NoneType
            or instance of Pydantic.BaseModel
        :param status_code: should be int
        """
        self.data = self._clean_data_type(data)
        self._check_status_code(status_code)
        self._headers = headers

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        else:
            return json.dumps(self.data)

    @property
    def headers(self) -> dict:
        return {
            'content-type': self.content_type,
            'access-control-allow-origin': '*',
        } | (self._headers or {})

    def _clean_data_type(self, data: any):
        """Make sure the response data is only ResponseDataTypes or Iterable of ResponseDataTypes"""
        if issubclass(type(data), PydanticBaseModel):
            return data.model_dump()

        elif isinstance(data, IterableDataTypes):
            return [self._clean_data_type(d) for d in data]

        elif isinstance(data, dict):
            return {key: self._clean_data_type(value) for key, value in data.items()}

        elif isinstance(data, (int | str | bool | bytes | NoneType)):
            return data

        else:
            msg = f'Invalid Response Type: {type(data)}'
            raise TypeError(msg)

    def _check_status_code(self, status_code: any):
        if not isinstance(status_code, int):
            error = f'Response "status_code" Should Be "int". ("{status_code}" is {type(status_code)})'
            raise TypeError(error)

        self.status_code = status_code

    def _clean_data_with_output_model(self, output_model: ModelMetaclass | None):
        if self.data and output_model:
            self.data = self._serialize_with_output_model(self.data, output_model=output_model)

    @classmethod
    def _serialize_with_output_model(cls, data: any, /, output_model: ModelMetaclass):
        # Dict
        if isinstance(data, dict):
            return output_model(**data).model_dump()

        # Iterable
        if isinstance(data, IterableDataTypes):
            return [cls._serialize_with_output_model(d, output_model=output_model) for d in data]

        # Str | Bool | Bytes
        msg = 'Type of Response data is not match with `output_model`.\n*hint: You may want to remove `output_model`'
        raise TypeError(msg)


class HTMLResponse(Response):
    content_type = 'text/html; charset=utf-8'

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        return self.data.encode()


class PlainTextResponse(Response):
    content_type = 'text/plain; charset=utf-8'

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        return self.data.encode()



import os
import sys

from panther import version as panther_version
from panther.cli.create_command import create
from panther.cli.monitor_command import monitor
from panther.cli.run_command import run
from panther.cli.utils import clean_args, cli_error, cli_info, cli_warning, print_help_message


def shell(args: list) -> None:
    if len(args) == 0:
        cli_info('You may want to use "bpython" or "ipython" for better interactive shell')
        os.system('python')
    elif len(args) != 1:
        return cli_error('Too Many Arguments.')
    shell_type = args[0].lower()
    if shell_type not in ['ipython', 'bpython']:
        return cli_error(f'"{shell_type}" Is Not Supported.')

    # Bpython
    if shell_type == 'bpython':
        try:
            import bpython
            os.system('bpython')
        except ImportError as e:
            cli_warning(e, 'Hint: "pip install ipython"')
            os.system('python')

    # Ipython
    elif shell_type == 'ipython':
        try:
            import IPython
            os.system('ipython')
        except ImportError as e:
            cli_warning(e, 'Hint: "pip install bpython"')
            os.system('python')


def version() -> None:
    print(panther_version())


def start() -> None:
    if len(sys.argv) < 2:
        cli_error('Please pass some arguments to the command.')
    else:
        command = sys.argv and sys.argv[1] or None
        args = clean_args(sys.argv[2:])

        match command:
            case '-h' | '--help':
                print_help_message()
            case 'create':
                create(sys.argv[2:])
            case 'run':
                run(args)
            case 'shell':
                shell(sys.argv[2:])
            case 'monitor':
                monitor()
            case 'version':
                version()
            case _:
                cli_error('Invalid Arguments.')



from unittest import TestCase

import bson
import faker
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, field_validator

from panther.configs import config
from panther.db.queries.mongodb_queries import BaseMongoDBQuery

f = faker.Faker()


class Model(PydanticBaseModel):
    id: str | None = Field(None, validation_alias='_id')

    @field_validator('id', mode='before')
    def validate_id(cls, value):
        if isinstance(value, str):
            try:
                bson.ObjectId(value)
            except bson.objectid.InvalidId:
                msg = 'Invalid ObjectId'
                raise ValueError(msg)
        elif not isinstance(value, bson.ObjectId):
            msg = 'ObjectId required'
            raise TypeError(msg)
        return str(value)

    @property
    def _id(self):
        return bson.ObjectId(self.id) if self.id else None


class Book(Model, BaseMongoDBQuery):
    name: str
    author: str
    pages_count: int


class TestPantherDB(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        config['db_engine'] = 'mongodb'

    @classmethod
    def tearDownClass(cls) -> None:
        config['db_engine'] = ''



import os
from datetime import datetime, timedelta
from pathlib import Path
from unittest import TestCase

from panther.utils import generate_hash_value_from_string, load_env, round_datetime


class TestLoadEnvFile(TestCase):
    file_path = 'tests/env-test'

    is_active = True
    db_host = '127.0.0.1'
    db_port = 27017

    def tearDown(self) -> None:
        Path(self.file_path).unlink()

    def _create_env_file(self, file_data):
        with open(self.file_path, 'w') as file:
            file.write(file_data)

    def test_load_env_double_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = "{self.is_active}"
DB_HOST = "{self.db_host}"
DB_PORT = "{self.db_port}"
        """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_single_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = '{self.is_active}'
DB_HOST = '{self.db_host}'
DB_PORT = '{self.db_port}'
                """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_no_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = {self.is_active}
DB_HOST = {self.db_host}
DB_PORT = {self.db_port}

                    """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_no_space(self):
        self._create_env_file(f"""
IS_ACTIVE={self.is_active}
DB_HOST={self.db_host}
DB_PORT={self.db_port}
                    """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_not_striped(self):
        self._create_env_file(f"""
        IS_ACTIVE = {self.is_active}
        DB_HOST = {self.db_host}
        DB_PORT = {self.db_port}
                    """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_and_read_from_system_env(self):
        self._create_env_file(f"""
IS_ACTIVE = '{self.is_active}'
DB_HOST = '{self.db_host}'
DB_PORT = '{self.db_port}'
                """)

        load_env(self.file_path)
        self.assertEqual(os.environ['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(os.environ['DB_HOST'], self.db_host)
        self.assertEqual(os.environ['DB_PORT'], str(self.db_port))


class TestUtilFunctions(TestCase):

    def test_round_datetime_second_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(seconds=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=40)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_second_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=35)
        _delta = timedelta(seconds=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=40)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_minute_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=15, second=30)
        _delta = timedelta(minutes=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=20)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_minute_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(minutes=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=20)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_hour_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(hours=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_hour_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=10, minute=10, second=30)
        _delta = timedelta(hours=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_day_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=10, minute=10, second=30)
        _delta = timedelta(days=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=23)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_day_2(self):
        _datetime = datetime(year=1997, month=12, day=22, hour=10, minute=10, second=30)
        _delta = timedelta(days=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=23)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_generate_hash_value_from_string(self):
        text = 'Hello World'
        hashed_1 = generate_hash_value_from_string(text)
        hashed_2 = generate_hash_value_from_string(text)

        self.assertEqual(hashed_1, hashed_2)
        self.assertNotEqual(text, hashed_1)



from collections import namedtuple
from collections.abc import Callable
from dataclasses import dataclass


@dataclass(frozen=True)
class Headers:
    accept_encoding: str
    content_length: int
    authorization: str
    content_type: str
    user_agent: str
    connection: str
    accept: str
    host: str

    sec_websocket_version: int
    sec_websocket_key: str
    upgrade: str


Address = namedtuple('Client', ['ip', 'port'])


class BaseRequest:
    def __init__(self, scope: dict, receive: Callable, send: Callable):
        self.scope = scope
        self.asgi_send = send
        self.asgi_receive = receive
        self._data = ...
        self._validated_data = None
        self._user = None
        self._headers: Headers | None = None
        self._params: dict | None = None

    @property
    def headers(self) -> Headers:
        _headers = {header[0].decode('utf-8'): header[1].decode('utf-8') for header in self.scope['headers']}
        if self._headers is None:
            self._headers = Headers(
                accept_encoding=_headers.pop('accept-encoding', None),
                content_length=_headers.pop('content_length', None),
                authorization=_headers.pop('authorization', None),
                content_type=_headers.pop('content-type', None),
                user_agent=_headers.pop('user-agent', None),
                connection=_headers.pop('connection', None),
                accept=_headers.pop('accept', None),
                host=_headers.pop('host', None),
                sec_websocket_version=_headers.pop('sec_websocket_version', None),
                sec_websocket_key=_headers.pop('sec_websocket_key', None),
                upgrade=_headers.pop('upgrade', None),
            )
        return self._headers

    @property
    def query_params(self) -> dict:
        if self._params is None:
            self._params = {}
            if (query_string := self.scope['query_string']) != b'':
                query_string = query_string.decode('utf-8').split('&')
                for param in query_string:
                    k, *_, v = param.split('=')
                    self._params[k] = v
        return self._params

    @property
    def path(self) -> str:
        return self.scope['path']

    @property
    def server(self) -> Address:
        return Address(*self.scope['server'])

    @property
    def client(self) -> Address:
        return Address(*self.scope['client'])

    @property
    def http_version(self) -> str:
        return self.scope['http_version']

    @property
    def scheme(self) -> str:
        return self.scope['scheme']

    @property
    def user(self):
        return self._user

    def set_user(self, user) -> None:
        self._user = user



import asyncio
import importlib
import logging
import re
from collections.abc import Callable
from traceback import TracebackException
from uuid import uuid4

import orjson as json

from panther import status
from panther.file_handler import File


logger = logging.getLogger('panther')


async def _http_response_start(send: Callable, /, headers: dict, status_code: int) -> None:
    bytes_headers = [[k.encode(), v.encode()] for k, v in (headers or {}).items()]
    await send({
        'type': 'http.response.start',
        'status': status_code,
        'headers': bytes_headers,
    })


async def _http_response_body(send: Callable, /, body: bytes | None = None) -> None:
    if body is None:
        await send({'type': 'http.response.body'})
    else:
        await send({'type': 'http.response.body', 'body': body})


async def http_response(
        send: Callable,
        /,
        *,
        status_code: int,
        monitoring=None,  # type: MonitoringMiddleware | None
        headers: dict | None = None,
        body: bytes | None = None,
        exception: bool = False,
) -> None:
    if exception:
        body = json.dumps({'detail': status.status_text[status_code]})
    elif status_code == status.HTTP_204_NO_CONTENT or body == b'null':
        body = None

    await monitoring.after(status_code)

    await _http_response_start(send, headers=headers, status_code=status_code)
    await _http_response_body(send, body=body)


def import_class(dotted_path: str, /) -> type:
    """
    Example:
    -------
        Input: panther.db.models.User
        Output: User (The Class)
    """
    path, name = dotted_path.rsplit('.', 1)
    module = importlib.import_module(path)
    return getattr(module, name)


def read_multipart_form_data(boundary: str, body: bytes) -> dict:
    r"""
    ----------------------------449529189836774544725855
    \r\nContent-Disposition: form-data; name="name"\r\n\r\nali\r\n
    ----------------------------449529189836774544725855
    \r\nContent-Disposition: form-data; name="image"; filename="ali.txt"\r\nContent-Type: text/plain\r\n\r\nHello\n\r\n
    ----------------------------449529189836774544725855
    \r\nContent-Disposition: form-data; name="age"\r\n\r\n12\r\n
    ----------------------------449529189836774544725855
    --\r\n
    """
    boundary = b'--' + boundary.encode()
    new_line = b'\r\n' if body[-2:] == b'\r\n' else b'\n'

    field_pattern = (
            rb'(Content-Disposition: form-data; name=")(.*)("'
            + 2 * new_line
            + b')(.*)'
    )
    file_pattern = (
            rb'(Content-Disposition: form-data; name=")(.*)("; filename=")(.*)("'
            + new_line
            + b'Content-Type: )(.*)'
    )

    data = {}
    for _row in body.split(boundary):
        row = _row.removeprefix(new_line).removesuffix(new_line)

        if row in (b'', b'--'):
            continue

        if match := re.match(pattern=field_pattern, string=row):
            _, field_name, _, value = match.groups()
            data[field_name.decode('utf-8')] = value.decode('utf-8')

        else:
            file_meta_data, value = row.split(2 * new_line, 1)

            if match := re.match(pattern=file_pattern, string=file_meta_data):
                _, field_name, _, file_name, _, content_type = match.groups()
                file = File(
                    file_name=file_name.decode('utf-8'),
                    content_type=content_type.decode('utf-8'),
                    file=value,
                )
                data[field_name.decode('utf-8')] = file
            else:
                logger.error('Unrecognized Pattern')

    return data


def generate_ws_connection_id() -> str:
    return uuid4().hex


def is_function_async(func: Callable) -> bool:
    """
    Sync result is 0 --> False
    async result is 128 --> True
    """
    return bool(func.__code__.co_flags & (1 << 7))


def run_sync_async_function(func: Callable):
    # Async
    if is_function_async(func):
        # Get Event Loop
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        # Add Coroutine To Event Loop
        if loop and loop.is_running():
            loop.create_task(func())

        # Start New Event Loop
        else:
            asyncio.run(func())

    # Sync
    else:
        func()


def clean_traceback_message(exception: Exception) -> str:
    """We are ignoring packages traceback message"""
    tb = TracebackException(type(exception), exception, exception.__traceback__)
    stack = tb.stack.copy()
    for t in stack:
        if t.filename.find('site-packages') != -1:
            tb.stack.remove(t)
    _traceback = list(tb.format(chain=False))
    return exception if len(_traceback) == 1 else f'{exception}\n' + ''.join(_traceback)



import time
from abc import abstractmethod
import logging
from typing import Literal

from jose import JWTError, jwt

from panther.configs import config
from panther.db.models import BaseUser, IDType, Model
from panther.exceptions import AuthenticationException
from panther.request import Request


logger = logging.getLogger('panther')


class BaseAuthentication:
    @classmethod
    @abstractmethod
    def authentication(cls, request: Request):
        """Return Instance of User"""
        msg = f'{cls.__name__}.authentication() is not implemented.'
        raise cls.exception(msg) from None

    @staticmethod
    def exception(message: str, /) -> type[AuthenticationException]:
        logger.error(f'Authentication Error: "{message}"')
        return AuthenticationException


class JWTAuthentication(BaseAuthentication):
    model = BaseUser
    keyword = 'Bearer'
    algorithm = 'HS256'
    HTTP_HEADER_ENCODING = 'iso-8859-1'  # Header encoding (see RFC5987)

    @classmethod
    def get_authorization_header(cls, request: Request) -> str:
        auth = request.headers.authorization

        if auth is None:
            msg = 'Authorization is required'
            raise cls.exception(msg) from None

        if isinstance(auth, str):
            auth = auth.encode(JWTAuthentication.HTTP_HEADER_ENCODING)
        return auth

    @classmethod
    def authentication(cls, request: Request) -> Model:
        auth = cls.get_authorization_header(request).split()
        if not auth or auth[0].lower() != cls.keyword.lower().encode():
            msg = 'Authorization keyword is not valid'
            raise cls.exception(msg) from None
        if len(auth) != 2:
            msg = 'Authorization should have 2 part'
            raise cls.exception(msg) from None

        try:
            token = auth[1].decode()
        except UnicodeError:
            msg = 'Unicode Error'
            raise cls.exception(msg) from None

        payload = cls.decode_jwt(token)
        return cls.get_user(payload)

    @classmethod
    def get_user(cls, payload: dict) -> Model:
        """Get UserModel from config, else use default UserModel from cls.model"""
        if (user_id := payload.get('user_id')) is None:
            msg = 'Payload does not have user_id'
            raise cls.exception(msg)

        user_model = config['user_model'] or cls.model
        if user := user_model.find_one(id=user_id):
            return user

        msg = 'User not found'
        raise cls.exception(msg) from None

    @classmethod
    def encode_jwt(cls, user_id: IDType, token_type: Literal['access', 'refresh'] = 'access') -> str:
        """Encode JWT from user_id."""
        issued_at = time.time()
        if token_type == 'access':
            expire = issued_at + config['jwt_config'].life_time
        else:
            expire = issued_at + config['jwt_config'].refresh_life_time

        claims = {
            'token_type': token_type,
            'user_id': user_id,
            'iat': issued_at,
            'exp': expire,
        }
        return jwt.encode(
            claims,
            key=config['jwt_config'].key,
            algorithm=config['jwt_config'].algorithm,
        )

    @classmethod
    def encode_refresh_token(cls, user_id: IDType) -> str:
        """Encode JWT from user_id."""
        return cls.encode_jwt(user_id=user_id, token_type='refresh')

    @classmethod
    def decode_jwt(cls, token: str) -> dict:
        """Decode JWT token to user_id (it can return multiple variable ... )"""
        try:
            return jwt.decode(
                token=token,
                key=config['jwt_config'].key,
                algorithms=[config['jwt_config'].algorithm],
            )
        except JWTError as e:
            raise cls.exception(e) from None

    @classmethod
    def login(cls, user_id: IDType) -> str:
        """Alias of encode_jwt()"""
        return cls.encode_jwt(user_id=user_id)

    @staticmethod
    def exception(message: str | JWTError, /) -> type[AuthenticationException]:
        logger.error(f'JWT Authentication Error: "{message}"')
        return AuthenticationException



from sys import version_info

from panther.db.connection import db
from panther.db.utils import merge_dicts, prepare_id_for_query
from panther.exceptions import DBException

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BaseMongoDBQuery')


class BaseMongoDBQuery:
    @classmethod
    def _merge(cls, *args) -> dict:
        prepare_id_for_query(*args, is_mongo=True)
        return merge_dicts(*args)

    # # # # # Find # # # # #
    @classmethod
    def find_one(cls, _data: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session[cls.__name__].find_one(cls._merge(_data, kwargs)):
            return cls.create_model_instance(document=document)
        return None

    @classmethod
    def find(cls, _data: dict | None = None, /, **kwargs) -> list[Self]:
        documents = db.session[cls.__name__].find(cls._merge(_data, kwargs))
        return [cls.create_model_instance(document=document) for document in documents]

    @classmethod
    def first(cls, _data: dict | None = None, /, **kwargs) -> Self | None:
        return cls.find_one(_data, **kwargs)

    @classmethod
    def last(cls, _data: dict | None = None, /, **kwargs):
        msg = 'last() is not supported in MongoDB yet.'
        raise DBException(msg)

    # # # # # Count # # # # #
    @classmethod
    def count(cls, _data: dict | None = None, /, **kwargs) -> int:
        return db.session[cls.__name__].count_documents(cls._merge(_data, kwargs))

    # # # # # Insert # # # # #
    @classmethod
    def insert_one(cls, _data: dict | None = None, **kwargs) -> Self:
        document = cls._merge(_data, kwargs)
        document['id'] = db.session[cls.__name__].insert_one(document).inserted_id
        return cls.create_model_instance(document=document)

    # # # # # Delete # # # # #
    def delete(self) -> None:
        db.session[self.__class__.__name__].delete_one({'_id': self._id})

    @classmethod
    def delete_one(cls, _data: dict | None = None, /, **kwargs) -> bool:
        result = db.session[cls.__name__].delete_one(cls._merge(_data, kwargs))
        return bool(result.deleted_count)

    @classmethod
    def delete_many(cls, _data: dict | None = None, /, **kwargs) -> int:
        result = db.session[cls.__name__].delete_many(cls._merge(_data, kwargs))
        return result.deleted_count

    # # # # # Update # # # # #
    def update(self, **kwargs) -> None:
        for field, value in kwargs.items():
            setattr(self, field, value)
        update_fields = {'$set': kwargs}
        db.session[self.__class__.__name__].update_one({'_id': self._id}, update_fields)

    @classmethod
    def update_one(cls, _filter: dict, _data: dict | None = None, /, **kwargs) -> bool:
        prepare_id_for_query(_filter, is_mongo=True)
        update_fields = {'$set': cls._merge(_data, kwargs)}

        result = db.session[cls.__name__].update_one(_filter, update_fields)
        return bool(result.matched_count)

    @classmethod
    def update_many(cls, _filter: dict, _data: dict | None = None, /, **kwargs) -> int:
        prepare_id_for_query(_filter, is_mongo=True)
        update_fields = {'$set': cls._merge(_data, kwargs)}

        result = db.session[cls.__name__].update_many(_filter, update_fields)
        return result.updated_count



import logging
import os
import hashlib
from datetime import datetime, timedelta
from pathlib import Path
from typing import ClassVar


logger = logging.getLogger('panther')


class Singleton(object):
    _instances: ClassVar[dict] = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__new__(cls, *args, **kwargs)
        return cls._instances[cls]


def load_env(env_file: str | Path, /) -> dict[str, str]:
    variables = {}

    if env_file is None or not Path(env_file).is_file():
        logger.critical(f'"{env_file}" is not valid file for load_env()')
        return variables

    with open(env_file) as file:
        for line in file.readlines():
            striped_line = line.strip()
            if not striped_line.startswith('#') and '=' in striped_line:
                key, value = striped_line.split('=', 1)
                key = key.strip()
                value = value.strip().strip('"\'')
                variables[key] = value

                # Load them as system environment variable
                os.environ[key] = value
    return variables


def generate_secret_key() -> str:
    from cryptography.fernet import Fernet

    return Fernet.generate_key().decode()


def round_datetime(dt: datetime, delta: timedelta):
    return datetime.min + round((dt - datetime.min) / delta) * delta


def generate_hash_value_from_string(string_value: str, /) -> str:
    # The point of this method is for maintenance, if we want to change
    # the hash algorithm in the future, it will be easy.
    return hashlib.sha256(string_value.encode('utf-8')).hexdigest()


def encrypt_password(password: str) -> str:
    return generate_hash_value_from_string(password)



import logging
from redis import Redis

from panther.db.connection import RedisConnection
from panther.middlewares.base import BaseMiddleware
from panther.request import Request
from panther.response import Response
from panther.websocket import GenericWebsocket


logger = logging.getLogger('panther')


class RedisMiddleware(BaseMiddleware):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.validate_host()
        self.validate_port()

    def validate_host(self):
        if host := self.kwargs.get('host'):  # noqa: F841
            if False:  # TODO: should be valid ip or domain (then remove the 'noqa')
                logger.critical('Redis "host" is not valid.')
        else:
            self.kwargs['host'] = '127.0.0.1'

    def validate_port(self):
        port = self.kwargs.setdefault('port', 6379)
        if not isinstance(port, int):
            logger.critical('Redis "port" is not valid.')

    async def before(self, request: Request | GenericWebsocket) -> Request | GenericWebsocket:
        self.redis = RedisConnection(**self.kwargs)
        return request

    async def after(self, response: Response | GenericWebsocket) -> Response | GenericWebsocket:
        self.redis.close()
        return response

    def redis_connection_for_ws(self) -> Redis:
        r = Redis(**self.kwargs)
        r.ping()
        return r



import sys
from unittest import TestCase

from panther import Panther
from panther.test import APIClient


class TestMethods(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/app')
        from tests.app.urls import method_urls

        app = Panther(__name__, configs=__name__, urls=method_urls)
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        sys.path.pop()

    def test_all(self):
        res = self.client.get('all/')
        assert res.status_code == 200

        res = self.client.post('all/')
        assert res.status_code == 200

        res = self.client.put('all/')
        assert res.status_code == 200

        res = self.client.patch('all/')
        assert res.status_code == 200

        res = self.client.delete('all/')
        assert res.status_code == 200

    def test_get(self):
        res = self.client.get('get/')
        assert res.status_code == 200

        res = self.client.post('get/')
        assert res.status_code == 405

        res = self.client.put('get/')
        assert res.status_code == 405

        res = self.client.patch('get/')
        assert res.status_code == 405

        res = self.client.delete('get/')
        assert res.status_code == 405

    def test_post(self):
        res = self.client.get('post/')
        assert res.status_code == 405

        res = self.client.post('post/')
        assert res.status_code == 200

        res = self.client.put('post/')
        assert res.status_code == 405

        res = self.client.patch('post/')
        assert res.status_code == 405

        res = self.client.delete('post/')
        assert res.status_code == 405

    def test_put(self):
        res = self.client.get('put/')
        assert res.status_code == 405

        res = self.client.post('put/')
        assert res.status_code == 405

        res = self.client.put('put/')
        assert res.status_code == 200

        res = self.client.patch('put/')
        assert res.status_code == 405

        res = self.client.delete('put/')
        assert res.status_code == 405

    def test_patch(self):
        res = self.client.get('patch/')
        assert res.status_code == 405

        res = self.client.post('patch/')
        assert res.status_code == 405

        res = self.client.put('patch/')
        assert res.status_code == 405

        res = self.client.patch('patch/')
        assert res.status_code == 200

        res = self.client.delete('patch/')
        assert res.status_code == 405

    def test_delete(self):
        res = self.client.get('delete/')
        assert res.status_code == 405

        res = self.client.post('delete/')
        assert res.status_code == 405

        res = self.client.put('delete/')
        assert res.status_code == 405

        res = self.client.patch('delete/')
        assert res.status_code == 405

        res = self.client.delete('delete/')
        assert res.status_code == 200

    def test_get_post_patch(self):
        res = self.client.get('get-post-patch/')
        assert res.status_code == 200

        res = self.client.post('get-post-patch/')
        assert res.status_code == 200

        res = self.client.put('get-post-patch/')
        assert res.status_code == 405

        res = self.client.patch('get-post-patch/')
        assert res.status_code == 200

        res = self.client.delete('get-post-patch/')
        assert res.status_code == 405



import sys
from unittest import TestCase

from panther import Panther
from panther.test import APIClient


class TestSimpleResponses(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/app')
        from tests.app.urls import simple_responses_urls

        app = Panther(__name__, configs=__name__, urls=simple_responses_urls)
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        sys.path.pop()

    def test_nothing(self):
        res = self.client.get('nothing/')
        assert res.status_code == 200
        assert res.data is None

    def test_none(self):
        res = self.client.get('none/')
        assert res.status_code == 200
        assert res.data is None

    def test_dict(self):
        res = self.client.get('dict/')
        assert res.status_code == 200
        assert res.data == {'detail': 'ok'}

    def test_list(self):
        res = self.client.get('list/')
        assert res.status_code == 200
        assert res.data == [1, 2, 3]

    def test_tuple(self):
        res = self.client.get('tuple/')
        assert res.status_code == 200
        assert res.data == [1, 2, 3, 4]

    def test_response_none(self):
        res = self.client.get('response-none/')
        assert res.status_code == 200
        assert res.data is None

    def test_response_dict(self):
        res = self.client.get('response-dict/')
        assert res.status_code == 200
        assert res.data == {'detail': 'ok'}

    def test_response_list(self):
        res = self.client.get('response-list/')
        assert res.status_code == 200
        assert res.data == ['car', 'home', 'phone']

    def test_response_tuple(self):
        res = self.client.get('response-tuple/')
        assert res.status_code == 200
        assert res.data == ['car', 'home', 'phone', 'book']



import logging
from typing import Literal

import orjson as json

from panther._utils import read_multipart_form_data
from panther.base_request import BaseRequest


logger = logging.getLogger('panther')


class Request(BaseRequest):
    @property
    def method(self) -> Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE']:
        return self.scope['method']

    @property
    def data(self) -> dict | bytes:
        """Data before validation"""
        if self._data is ...:
            match (self.headers.content_type or '').split('; boundary='):
                case ['' | 'application/json']:
                    self._data = json.loads(self.__body or b'{}')
                case ['multipart/form-data', boundary]:
                    self._data = read_multipart_form_data(boundary=boundary, body=self.__body)
                case [unknown]:
                    # We don't know the `content-type` so just pass the payload to user
                    logger.warning(f"'{unknown}' Content-Type is not supported")
                    self._data = self.__body
        return self._data

    def set_validated_data(self, validated_data) -> None:
        self._validated_data = validated_data

    @property
    def validated_data(self):
        """
        Return The Validated Data
        It has been set on API.validate_input() while request is happening.
        """
        return getattr(self, '_validated_data', None)

    async def read_body(self) -> None:
        """Read and return the entire body from an incoming ASGI message."""
        self.__body = b''
        more_body = True
        while more_body:
            message = await self.asgi_receive()
            self.__body += message.get('body', b'')
            more_body = message.get('more_body', False)



import sys
from pathlib import Path
from unittest import TestCase

from panther import Panther


class TestRun(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/run')

    @classmethod
    def tearDownClass(cls) -> None:
        sys.path.pop()

    def test_init(self):
        app = Panther(__name__)
        self.assertIsInstance(app, Panther)

    def test_load_configs(self):
        from panther.configs import config
        from panther.db.models import BaseUser
        from panther.panel.apis import documents_api, models_api, single_document_api

        base_dir = Path(__name__).resolve().parent
        app = Panther(__name__)

        self.assertIsInstance(app.configs, dict)
        self.assertEqual(app.configs.get('URLs'), 'core.configs.urls')

        self.assertIsInstance(config, dict)
        self.assertEqual(config['base_dir'], base_dir)
        self.assertFalse(config['monitoring'])
        self.assertFalse(config['log_queries'])
        self.assertFalse(config['default_cache_exp'])
        self.assertIsNone(config['throttling'])
        self.assertIsNone(config['secret_key'])
        self.assertEqual(config['middlewares'], [])
        self.assertEqual(config['reversed_middlewares'], [])
        self.assertEqual(config['user_model'], BaseUser)
        self.assertIsNone(config['jwt_config'])
        self.assertEqual(len(config['models']), 2)

        self.assertTrue('' in config['urls'])
        config['urls'].pop('')
        urls = {
            '_panel': {
                '': models_api,
                '<index>': {
                    '': documents_api,
                    '<document_id>': single_document_api,
                },
            },
        }
        self.assertEqual(config['urls'], urls)
        self.assertEqual(config['db_engine'], '')



import ast
import platform
import sys
from datetime import timedelta
from importlib import import_module
from pathlib import Path
from typing import Callable

from pydantic._internal._model_construction import ModelMetaclass

from panther._utils import import_class
from panther.configs import JWTConfig, config
from panther.exceptions import PantherException
from panther.routings import finalize_urls, flatten_urls
from panther.throttling import Throttling

__all__ = (
    'load_configs_file',
    'load_secret_key',
    'load_monitoring',
    'load_log_queries',
    'load_background_tasks',
    'load_throttling',
    'load_default_cache_exp',
    'load_middlewares',
    'load_user_model',
    'load_authentication_class',
    'load_jwt_config',
    'load_startup',
    'load_shutdown',
    'collect_all_models',
    'load_urls',
    'load_panel_urls',
)


def load_configs_file(_configs, /) -> dict:
    """Read the config file and put it as dict in self.configs"""
    if _configs:
        _module = sys.modules[_configs]
    else:
        try:
            _module = import_module('core.configs')
        except ModuleNotFoundError:
            raise _exception_handler(field='core/configs.py', error='Not Found')
    return _module.__dict__


def load_secret_key(configs: dict, /) -> bytes | None:
    if secret_key := configs.get('SECRET_KEY'):
        return secret_key.encode()
    return secret_key


def load_monitoring(configs: dict, /) -> bool:
    return configs.get('MONITORING', config['monitoring'])


def load_log_queries(configs: dict, /) -> bool:
    return configs.get('LOG_QUERIES', config['log_queries'])


def load_background_tasks(configs: dict, /) -> bool:
    return configs.get('BACKGROUND_TASKS', config['background_tasks'])


def load_throttling(configs: dict, /) -> Throttling | None:
    return configs.get('THROTTLING', config['throttling'])


def load_default_cache_exp(configs: dict, /) -> timedelta | None:
    return configs.get('DEFAULT_CACHE_EXP', config['default_cache_exp'])


def load_middlewares(configs: dict, /) -> list:
    """Collect The Middlewares & Set db_engine If One Of Middlewares Was For DB"""
    from panther.middlewares import BaseMiddleware

    middlewares = []

    for path, data in configs.get('MIDDLEWARES', []):
        if path.find('panther.middlewares.db.DatabaseMiddleware') != -1:
            config['db_engine'] = data['url'].split(':')[0]
        try:
            Middleware = import_class(path)  # noqa: N806
        except AttributeError:
            raise _exception_handler(field='MIDDLEWARES', error=f'{path} is not a valid middleware path')

        if not issubclass(Middleware, BaseMiddleware):
            raise _exception_handler(field='MIDDLEWARES', error='is not a sub class of BaseMiddleware')

        middlewares.append(Middleware(**data))  # noqa: Py Argument List
    return middlewares


def load_user_model(configs: dict, /) -> ModelMetaclass:
    return import_class(configs.get('USER_MODEL', 'panther.db.models.BaseUser'))


def load_authentication_class(configs: dict, /) -> ModelMetaclass | None:
    return configs.get('AUTHENTICATION') and import_class(configs['AUTHENTICATION'])


def load_jwt_config(configs: dict, /) -> JWTConfig | None:
    """Only Collect JWT Config If Authentication Is JWTAuthentication"""
    if getattr(config['authentication'], '__name__', None) == 'JWTAuthentication':
        user_config = configs.get('JWTConfig', {})
        if 'key' not in user_config:
            user_config['key'] = config['secret_key'].decode()

        return JWTConfig(**user_config)
    return None


def load_startup(configs: dict, /) -> Callable:
    return configs.get('STARTUP') and import_class(configs['STARTUP'])


def load_shutdown(configs: dict, /) -> Callable:
    return configs.get('SHUTDOWN') and import_class(configs['SHUTDOWN'])


def collect_all_models() -> list[dict]:
    """Collecting all models for panel APIs"""
    from panther.db.models import Model

    # Just load all the python files from 'base_dir',
    #   so Model.__subclasses__ can find all the subclasses
    slash = '\\' if platform.system() == 'Windows' else '/'
    python_files = [
        f for f in config['base_dir'].rglob('*.py')
        if not f.name.startswith('_') and 'site-packages' not in f.parents._parts
    ]
    for file in python_files:
        # Analyse the file
        with Path(file).open() as f:
            node = ast.parse(f.read())

        model_imported = False
        panther_imported = False
        panther_called = False
        for n in node.body:
            match n:

                # from panther.db import Model
                case ast.ImportFrom(module='panther.db', names=[ast.alias(name='Model')]):
                    model_imported = True

                # from panther.db.models import ..., Model, ...
                case ast.ImportFrom(module='panther.db.models', names=[*names]):
                    try:
                        next(v for v in names if v.name == 'Model')
                        model_imported = True
                    except StopIteration:
                        pass

                # from panther import Panther, ...
                case ast.ImportFrom(module='panther', names=[ast.alias(name='Panther'), *_]):
                    panther_imported = True

                # from panther import ..., Panther
                case ast.ImportFrom(module='panther', names=[*_, ast.alias(name='Panther')]):
                    panther_imported = True

                # ... = Panther(...)
                case ast.Assign(value=ast.Call(func=ast.Name(id='Panther'))):
                    panther_called = True

        # Panther() should not be called in the file and Model() should be imported,
        #   We check the import of the Panther to make sure he is calling the panther.Panther and not any Panther
        if panther_imported and panther_called or not model_imported:
            continue

        # Load the module
        dotted_f = str(file).removeprefix(f'{config["base_dir"]}{slash}').removesuffix('.py').replace(slash, '.')
        import_module(dotted_f)

    return [
        {
            'name': m.__name__,
            'module': m.__module__,
            'class': m
        } for m in Model.__subclasses__() if m.__module__ != 'panther.db.models'
    ]


def load_urls(configs: dict, /, urls: dict | None) -> tuple[dict, dict]:
    """
    Return tuple of all urls (as a flat dict) and (as a nested dict)
    """
    if isinstance(urls, dict):
        collected_urls = flatten_urls(urls)
        return collected_urls, finalize_urls(collected_urls)

    if (url_routing := configs.get('URLs')) is None:
        raise _exception_handler(field='URLs', error='is required.')

    if isinstance(url_routing, dict):
        error = (
            "can't be 'dict', you may want to pass it's value directly to Panther(). " 'Example: Panther(..., urls=...)'
        )
        raise _exception_handler(field='URLs', error=error)

    if not isinstance(url_routing, str):
        error = 'should be dotted string.'
        raise _exception_handler(field='URLs', error=error)

    try:
        imported_urls = import_class(url_routing)
    except ModuleNotFoundError as e:
        raise _exception_handler(field='URLs', error=e)

    if not isinstance(imported_urls, dict):
        raise _exception_handler(field='URLs', error='should point to a dict.')

    collected_urls = flatten_urls(imported_urls)
    return collected_urls, finalize_urls(collected_urls)


def load_panel_urls() -> dict:
    from panther.panel.urls import urls

    return finalize_urls(flatten_urls(urls))


def _exception_handler(field: str, error: str | Exception) -> PantherException:
    return PantherException(f"Invalid '{field}': {error}")



from datetime import datetime

import bson
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, field_validator

from panther.configs import config
from panther.db.queries import Query

IDType = int if config['db_engine'] == 'pantherdb' else str


class Model(PydanticBaseModel, Query):
    id: IDType | None = Field(None, validation_alias='_id')

    @field_validator('id', mode='before')
    def validate_id(cls, value: IDType | bson.ObjectId) -> IDType:
        if IDType is str:
            if isinstance(value, str):
                try:
                    bson.ObjectId(value)
                except bson.objectid.InvalidId as e:
                    msg = 'Invalid ObjectId'
                    raise ValueError(msg) from e
            elif not isinstance(value, bson.ObjectId):
                msg = 'ObjectId required'
                raise ValueError(msg) from None
            value = str(value)
        return value

    @property
    def _id(self) -> int | bson.ObjectId | None:
        if IDType is int:
            return self.id
        return bson.ObjectId(self.id) if self.id else None

    def dict(self, *args, **kwargs) -> dict:
        return self.model_dump(*args, **kwargs)


class BaseUser(Model):
    first_name: str = Field('', max_length=64)
    last_name: str = Field('', max_length=64)
    last_login: datetime = None

    def update_last_login(self) -> None:
        self.update(last_login=datetime.now())



import asyncio
from collections.abc import Callable
from typing import Literal

import orjson as json

from panther.response import Response

__all__ = ('APIClient',)


class RequestClient:
    def __init__(self, app: Callable):
        self.app = app

    async def send(self, data: dict):
        if data['type'] == 'http.response.start':
            self.header = data
        else:
            self.response = data

    async def receive(self):
        return {
            'type': 'http.request',
            'body': self.payload,
            'more_body': False,
        }

    async def request(
            self,
            path: str,
            method: Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE'],
            payload: dict | None,
            headers: dict,
            query_params: dict,
    ) -> Response:
        headers = [(k.encode(), v.encode()) for k, v in headers.items()]
        if not path.startswith('/'):
            path = f'/{path}'

        self.payload = json.dumps(payload)
        query_params = '&'.join(f'{k}={v}' for k, v in query_params.items())
        scope = {
            'type': 'http',
            'client': ('127.0.0.1', 8585),
            'headers': headers,
            'method': method,
            'path': path,
            'raw_path': path.encode(),
            'query_string': query_params.encode(),
        }
        await self.app(
            scope=scope,
            receive=self.receive,
            send=self.send,
        )
        return Response(
            data=json.loads(self.response.get('body', b'null')),
            status_code=self.header['status'],
            headers=self.header['headers'],
        )


class APIClient:
    def __init__(self, app: Callable):
        self._app = app

    def _send_request(
            self,
            path: str,
            method: Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE'],
            payload: dict | None,
            headers: dict,
            query_params: dict,
    ) -> Response:
        request_client = RequestClient(app=self._app)
        return asyncio.run(
            request_client.request(
                path=path,
                method=method,
                payload=payload,
                headers=headers,
                query_params=query_params or {},
            )
        )

    def get(
            self,
            path: str,
            headers: dict | None = None,
            query_params: dict | None = None,
    ) -> Response:
        return self._send_request(
            path=path,
            method='GET',
            payload=None,
            headers=headers or {},
            query_params=query_params or {},
        )

    def post(
            self,
            path: str,
            payload: dict | None = None,
            headers: dict | None = None,
            query_params: dict | None = None,
            content_type: Literal['application/json', 'multipart/form-data'] = 'application/json',
    ) -> Response:
        headers = {'content-type': content_type} | (headers or {})
        return self._send_request(
            path=path,
            method='POST',
            payload=payload,
            headers=headers,
            query_params=query_params or {},
        )

    def put(
            self,
            path: str,
            payload: dict | None = None,
            headers: dict | None = None,
            query_params: dict | None = None,
            content_type: Literal['application/json', 'multipart/form-data'] = 'application/json',
    ) -> Response:
        headers = {'content-type': content_type} | (headers or {})
        return self._send_request(
            path=path,
            method='PUT',
            payload=payload,
            headers=headers,
            query_params=query_params or {},
        )

    def patch(
            self,
            path: str,
            payload: dict | None = None,
            headers: dict | None = None,
            query_params: dict | None = None,
            content_type: Literal['application/json', 'multipart/form-data'] = 'application/json',
    ) -> Response:
        headers = {'content-type': content_type} | (headers or {})
        return self._send_request(
            path=path,
            method='PATCH',
            payload=payload,
            headers=headers,
            query_params=query_params or {},
        )

    def delete(
            self,
            path: str,
            headers: dict | None = None,
            query_params: dict | None = None,
    ) -> Response:
        return self._send_request(
            path=path,
            method='DELETE',
            payload=None,
            headers=headers or {},
            query_params=query_params or {},
        )
